from equadratures.parameter import Parameter
from equadratures.poly import Poly
from equadratures.basis import Basis
from equadratures.scalers import scaler_minmax, scaler_meanvar, scaler_custom
import equadratures.plot as plot
import numpy as np
import scipy
import scipy.io
from scipy.linalg import orth, sqrtm
from scipy.spatial import ConvexHull
from scipy.special import comb
from scipy.optimize import linprog
import warnings

class Subspaces(object):
    """
    This class defines a subspaces object. It can be used for polynomial-based subspace dimension reduction.

    :param string method: The method to be used for subspace-based dimension reduction. One option is
    ``active-subspace``, which uses ideas in [1] and [2] to compute a dimension-reducing subspace with a global
     polynomial approximant. Gradients evaluations of the polynomial approximation are used to compute the averaged
      outer product of the gradient covariance matrix. Another option is ``variable-projection`` [3], where a
      Gauss-Newton optimisation problem is solved to compute both the polynomial coefficients and its subspace.
    :param numpy.ndarray sample_points: A numpy ndarray with shape (number_of_observations, dimensions) that corresponds
     to a set of sample points over the parameter space.
    :param numpy.ndarray sample_outputs: A numpy ndarray with shape (number_of_observations, 1) that corresponds to
    model evaluations at the sample points.
    :param int subspace_dimension: The dimension of the *active* subspace.
    :param dict param_args: Arguments passed to parameters of the AS polynomial. (see `parameter.py`)
    :param dict poly_args: Arguments passed to constructing polynomial used for AS computation. (see `poly.py`)
    :param dict dr_args: Arguments passed to customise the VP optimiser. See documentation for _get_variable_projection.

    **Sample constructor initialisations**::

        import numpy as np
        from equadratures import *

        # Active subspaces with a global order 2 polynomial.
        mysubspace = Subspaces(method='active-subspace', sample_points=X_red, sample_outputs=Y_red)
        eigs = mysubspace.get_eigenvalues()
        W = mysubspace.get_subspace()
        e = mysubspace.get_eigenvalues()

    **References**
        1. Constantine, P., (2015) Active Subspaces: Emerging Ideas for Dimension Reduction in Parameter Studies.
        SIAM Spotlights.
        2. Seshadri, P., Shahpar, S., Constantine, P., Parks, G., Adams, M. (2018) Turbomachinery Active Subspace
         Performance Maps. Journal of Turbomachinery, 140(4), 041003.
         `Paper <http://turbomachinery.asmedigitalcollection.asme.org/article.aspx?articleid=2668256>`__.
        3. Hokanson, J., Constantine, P., (2018) Data-driven Polynomial Ridge Approximation Using Variable Projection.
         SIAM Journal of Scientific Computing, 40(3), A1566-A1589.
          `Paper <https://epubs.siam.org/doi/abs/10.1137/17M1117690>`__.
    """
    def __init__(self, method, full_space_poly=None, sample_points=None, sample_outputs=None,
                 subspace_dimension=2, polynomial_degree=2, param_args=None, poly_args=None, dr_args=None):
        self.full_space_poly = full_space_poly
        self.sample_points = sample_points
        self.Y = None  # for the zonotope vertices

        '''
        How to standardise?
        If just a dataset, no poly, then use minmax scaling
        If poly, no datapoints, then data should be contained in poly
         - extract points from poly using .get_points() and .get_model_evaluations
         - standaridise according to distributions extracted from the poly's params
        If poly, but with new points? We should assume points are drawn from same distr according to poly's params
         - standaridise according to distributions extracted from the poly's params
        We need information on whether marginals are bounded, semi-bounded or unbounded?
         - bounded, minmax scaling: analytical, beta, chebyshev, truncated gaussian, uniform
         - unbounded, meanvar scaling: cauchy (although stddev undefined!), gaussian, gumbel, logistic, studentst,
         - semi-bounded, ???: chi, chi squared, exponential, gamma, lognormal, pareto (x \geq 1), rayleigh, weibull
        '''

        self.sample_outputs = sample_outputs
        self.method = method
        self.subspace_dimension = subspace_dimension
        self.polynomial_degree = polynomial_degree

        my_poly_args = {'method': 'least-squares', 'solver_args': {}}
        if poly_args is not None:
            my_poly_args.update(poly_args)
        self.poly_args = my_poly_args

        my_param_args = {'distribution': 'uniform', 'order': self.polynomial_degree, 'lower': -1, 'upper': 1}
        if param_args is not None:
            my_param_args.update(param_args)

        # I suppose we can detect if lower and upper is present to decide between these categories?
        bounded_distrs = ['analytical', 'beta', 'chebyshev', 'arcsine', 'truncated-gaussian', 'uniform']
        unbounded_distrs = ['gaussian', 'normal', 'gumbel', 'logistic', 'students-t', 'studentst']
        semi_bounded_distrs = ['chi', 'chi-squared', 'exponential', 'gamma', 'lognormal', 'log-normal', 'pareto', 'rayleigh', 'weibull']

        if self.method.lower() == 'active-subspace' or self.method.lower() == 'active-subspaces':
            self.method = 'active-subspace'
            if dr_args is not None:
                self.standardise = getattr(dr_args, 'standardise', True)
            else:
                self.standardise = True

            if self.full_space_poly is None:
                # user provided input/output data
                N, d = self.sample_points.shape
                if self.standardise:
                    self.data_scaler = scaler_minmax()
                    self.data_scaler.fit(self.sample_points)
                    self.std_sample_points = self.data_scaler.transform(self.sample_points)
                else:
                    self.std_sample_points = self.sample_points.copy()
                param = Parameter(**my_param_args)
                if param_args is not None:
                    if (hasattr(dr_args, 'lower') or hasattr(dr_args, 'upper')) and self.standardise:
                        warnings.warn('Points standardised but parameter range provided. Overriding default ([-1,1])...',
                            UserWarning)
                myparameters = [param for _ in range(d)]
                mybasis = Basis("total-order")
                mypoly = Poly(myparameters, mybasis, sampling_args={'sample-points': self.std_sample_points,
                                                                    'sample-outputs': self.sample_outputs},
                              **my_poly_args)
                mypoly.set_model()
                self.full_space_poly = mypoly
            else:
                # User provided polynomial
                # Standardise according to distribution specified. Only care about the scaling (not shift)
                # TODO: user provided callable with parameters?
                user_params = self.full_space_poly.parameters
                d = len(user_params)
                self.sample_points = self.full_space_poly.get_points()
                if self.standardise:
                    scale_factors = np.zeros(d)
                    centers = np.zeros(d)
                    for dd, p in enumerate(user_params):
                        if p.name.lower() in bounded_distrs:
                            print(p.name)
                            scale_factors[dd] = (p.upper - p.lower) / 2.0
                            centers[dd] = (p.upper + p.lower) / 2.0
                        elif p.name.lower() in unbounded_distrs:
                            scale_factors[dd] = np.sqrt(p.variance)
                            centers[dd] = p.mean
                        else:
                            scale_factors[dd] = np.sqrt(p.variance)
                            centers[dd] = 0.0
                    self.param_scaler = scaler_custom(centers, scale_factors)
                    self.std_sample_points = self.param_scaler.transform(self.sample_points)
                else:
                    self.std_sample_points = self.sample_points.copy()
                if not hasattr(self.full_space_poly, 'coefficients'):
                    raise ValueError('Please call set_model() first on poly.')

            self.sample_outputs = self.full_space_poly.get_model_evaluations()
            # TODO: use dr_args for resampling of gradient points
            as_args = {'grad_points': None}
            if dr_args is not None:
                as_args.update(dr_args)
            self._get_active_subspace(**as_args)
        elif self.method == 'variable-projection':
            self.data_scaler = scaler_minmax()
            self.data_scaler.fit(self.sample_points)
            self.std_sample_points = self.data_scaler.transform(self.sample_points)

            if dr_args is not None:
                vp_args = {'gamma':0.1, 'beta':1e-4, 'tol':1e-7, 'maxiter':1000, 'U0':None, 'verbose':False}
                vp_args.update(dr_args)
                self._get_variable_projection(**vp_args)
            else:
                self._get_variable_projection()

    def get_subspace_polynomial(self):
        """
        Returns a polynomial defined over the dimension reducing subspace.

        :param Subspaces self:
            An instance of the Subspaces object.

        :return:
            **subspacepoly**: A Poly object that defines a polynomial over the subspace. The distribution of parameters
             is assumed to be uniform and the maximum and minimum bounds for each parameter are defined by the maximum
              and minimum values of the project samples.
        """
        # TODO: Try correlated poly here
        active_subspace = self._subspace[:, 0:self.subspace_dimension]
        projected_points = np.dot(self.std_sample_points, active_subspace)
        myparameters = []
        for i in range(0, self.subspace_dimension):
            param = Parameter(distribution='uniform', lower=np.min(projected_points[:, i]),
                              upper=np.max(projected_points[:, i]), order=self.polynomial_degree)
            myparameters.append(param)
        mybasis = Basis("total-order")
        subspacepoly = Poly(myparameters, mybasis, method='least-squares',
                            sampling_args={'sample-points': projected_points,
                                           'sample-outputs': self.sample_outputs})
        subspacepoly.set_model()
        return subspacepoly

    def get_eigenvalues(self):
        """
        Returns the eigenvalues of the dimension reducing subspace. Note: this option is
        currently only valid for method ``active-subspace``.

        :param Subspaces self:
            An instance of the Subspaces object.

        :return:
            **eigenvalues**: A numpy.ndarray of shape (dimensions,) corresponding to the eigenvalues of the above mentioned covariance matrix.
        """
        if self.method == 'active-subspace':
            return self._eigenvalues
        else:
            print('Only the active-subspace method yields eigenvalues.')

    def get_subspace(self):
        """
        Returns the dimension reducing subspace.

        :param Subspaces self:
            An instance of the Subspaces object.

        :return:
            **subspace**: A numpy.ndarray of shape (dimensions, dimensions) where the first ``subspace_dimension`` columns
            contain the dimension reducing subspace, while the remaining columns contain its orthogonal complement.
        """
        return self._subspace

    def _get_active_subspace(self, grad_points=None, **kwargs):
        """
        Active subspaces.
        """
        if grad_points is None:
            X = self.full_space_poly.get_points()
        else:
            if hasattr(self, 'data_scaler'):
                X = self.data_scaler.transform(grad_points)
            else:
                # Either no standardisation, or user provided poly + param scaling
                X = grad_points.copy()

        M, d = X.shape
        if d != self.sample_points.shape[1]:
            raise ValueError('In _get_active_subspace: dimensions of gradient evaluation points mismatched with input dimension!')

        alpha = 2.0
        num_grad_lb = alpha * self.subspace_dimension * np.log(d)
        if M < num_grad_lb:
            warnings.warn('Number of gradient evaluation points is likely to be insufficient. Consider resampling!', UserWarning)

        polygrad = self.full_space_poly.get_polyfit_grad(X)
        if hasattr(self, 'param_scaler'):
            # Evaluate gradient in transformed coordinate space
            polygrad = self.param_scaler.ranges[:, np.newaxis] * polygrad
        weights = np.ones((M, 1)) / M
        R = polygrad.transpose() * weights
        C = np.dot(polygrad, R )

        # Compute eigendecomposition!
        e, W = np.linalg.eigh(C)
        idx = e.argsort()[::-1]
        eigs = e[idx]
        eigVecs = W[:, idx]
        if hasattr(self, 'data_scaler'):
            scale_factors = 2.0 / (self.data_scaler.Xmax - self.data_scaler.Xmin)
            eigVecs = scale_factors[:, np.newaxis] * eigVecs
            eigVecs = np.linalg.qr(eigVecs)[0]

        self._subspace = eigVecs
        self._eigenvalues = eigs

    def _get_variable_projection(self, gamma=0.1, beta=1e-4, tol=1e-7, maxiter=1000, U0=None, verbose=False, **kw_args):
        """
        Variable Projection function to obtain an active subspace in inputs design space
        Note: It may help to standardize outputs to zero mean and unit variance
        :param gamma: double, step length reduction factor (0,1)
        :param beta: double, Armijo tolerance for backtracking line search (0,1)
        :param tol: double, tolerance for convergence, measured in the norm of residual over norm of f
        :param maxiter: int, maximum number of optimisation iterations
        :param U0: numpy.ndarray, initial guess for active subspace
        :param verbose: bool, set to True for debug messages
        """
        # NOTE: How do we know these are the best values of gamma and beta?
        M, m = self.std_sample_points.shape
        if U0 is None:
            Z = np.random.randn(m, self.subspace_dimension)
            U, _ = np.linalg.qr(Z)
        else:
            U = orth(U0)
        y = np.dot(self.std_sample_points,U)
        minmax = np.zeros((2, self.subspace_dimension))
        minmax[0, :] = np.amin(y, axis=0)
        minmax[1, :] = np.amax(y, axis=0)
        # Construct the affine transformation
        eta = 2 * np.divide((y - minmax[0,:]), (minmax[1,:]-minmax[0,:])) - 1

        # Construct the Vandermonde matrix step 6

        V, poly_obj = vandermonde(eta, self.polynomial_degree)
        V_plus = np.linalg.pinv(V)
        coeff = np.dot(V_plus, self.sample_outputs)
        res = self.sample_outputs - np.dot(V,coeff)
        # R = np.linalg.norm(res)
        # TODO: convergence criterion??

        for iteration in range(0,maxiter):
            # Construct the Jacobian step 9
            J = jacobian_vp(V, V_plus, U, self.sample_outputs, poly_obj, eta, minmax, self.std_sample_points)
            # Calculate the gradient of Jacobian (step 10)
            G = np.zeros((m, self.subspace_dimension))
            # NOTE: Can be vectorised
            for i in range(0, M):
                G += res[i]*J[i, :, :]

            # conduct the SVD for J_vec
            vec_J = np.reshape(J, (M, m*self.subspace_dimension))
            Y, S, Z = np.linalg.svd(vec_J,full_matrices=False)  # step 11

            # obtain delta
            delta = np.dot(Y[:,:-self.subspace_dimension**2].T, res)
            delta = np.dot(np.diag(1/S[:-self.subspace_dimension**2]), delta)
            delta = -np.dot(Z[:-self.subspace_dimension**2,:].T, delta).reshape(U.shape)

            # carry out Gauss-Newton step
            vec_delta=delta.flatten()  # step 12

            # vectorize G step 13
            vec_G = G.flatten()
            alpha = np.dot(vec_G.T, vec_delta)
            norm_G = np.dot(vec_G.T, vec_G)

            # check alpha step 14
            if alpha >= 0:
                delta = -G
                alpha = -norm_G

            # SVD on delta step 17
            
            Y, S, Z = np.linalg.svd(delta, full_matrices=False)

            UZ = np.dot(U,Z.T)
            t = 1
            for iter2 in range(0,20):
                U_new = np.dot(UZ, np.diag(np.cos(S*t))) + np.dot(Y, np.diag(np.sin(S*t)))#step 19
                U_new = orth(U_new)
                # Update the values with the new U matrix
                y = np.dot(self.std_sample_points, U_new)
                minmax[0,:] = np.amin(y, axis=0)
                minmax[1,:] = np.amax(y, axis=0)
                eta = 2 * np.divide((y - minmax[0,:]), (minmax[1,:]-minmax[0,:])) - 1

                V_new, poly_obj = vandermonde(eta, self.polynomial_degree)
                V_plus_new = np.linalg.pinv(V_new)
                coeff_new = np.dot(V_plus_new, self.sample_outputs)
                res_new = self.sample_outputs - np.dot(V_new,coeff_new)
                R_new = np.linalg.norm(res_new)

                if np.linalg.norm(res_new) <= np.linalg.norm(res)+alpha*beta*t or t < 1e-10:  # step 21
                    break
                t = t * gamma
            dist_change = subspace_dist(U, U_new)
            U = U_new
            V = V_new
            # coeff = coeff_new
            V_plus = V_plus_new
            res = res_new
            # R = R_new
            if dist_change < tol:
                if verbose:
                    print("VP finished with %d iterations" % iteration)
                break
        if iteration == maxiter - 1 and verbose:
            print("VP finished with %d iterations" % iteration)
        active_subspace = U
        inactive_subspace = scipy.linalg.null_space(active_subspace.T)
        self._subspace = np.hstack([active_subspace, inactive_subspace])

    def get_zonotope_vertices(self, num_samples=10000, max_count=100000):
        """
        Returns the vertices of the zonotope -- the projection of the high-dimensional space over the computed
        subspace.

        :param Subspaces self:
            An instance of the Subspaces object.

        :return:
            **vertices**: A numpy.ndarray of shape (number of vertices, ``subspace_dimension``).

        **Note:**
        This routine has been adapted from Paul Constantine's zonotope_vertices() function; see reference below.

        Constantine, P., Howard, R., Glaws, A., Grey, Z., Diaz, P., Fletcher, L., (2016) Python Active-Subspaces Utility Library. Journal of Open Source Software, 1(5), 79. `Paper <http://joss.theoj.org/papers/10.21105/joss.00079>`__.
        """
        m = self._subspace.shape[0]
        n = self.subspace_dimension
        W = self._subspace[:, :n]
        if n == 1:
            y0 = np.dot(W.T, np.sign(W))[0]
            if y0 < -y0:
                yl, yu = y0, -y0
                xl, xu = np.sign(W), -np.sign(W)
            else:
                yl, yu = -y0, y0
                xl, xu = -np.sign(W), np.sign(W)
            Y = np.array([yl, yu]).reshape((2,1))
            X = np.vstack((xl.reshape((1,m)), xu.reshape((1,m))))
            self.Y = Y
            return Y
        else:
            total_vertices = 0
            for i in range(n):
                total_vertices += comb(m-1,i)
            total_vertices = int(2*total_vertices)

            Z = np.random.normal(size=(num_samples, n))
            X = get_unique_rows(np.sign(np.dot(Z, W.transpose())))
            X = get_unique_rows(np.vstack((X, -X)))
            N = X.shape[0]

            count = 0
            while N < total_vertices:
                Z = np.random.normal(size=(num_samples, n))
                X0 = get_unique_rows(np.sign(np.dot(Z, W.transpose())))
                X0 = get_unique_rows(np.vstack((X0, -X0)))
                X = get_unique_rows(np.vstack((X, X0)))
                N = X.shape[0]
                count += 1
                if count > max_count:
                    break

            num_vertices = X.shape[0]
            if total_vertices > num_vertices:
                print('Warning: {} of {} vertices found.'.format(num_vertices, total_vertices))

            Y = np.dot(X, W)
            self.Y = Y.reshape((num_vertices, n))
            return self.Y

    def get_linear_inequalities(self):
        """
        Returns the linear inequalities defining the zonotope vertices, i.e., Ax<=b.

        :param Subspaces self:
            An instance of the Subspaces object.

        :return:
            **A**: The matrix for setting the linear inequalities.
        :return:
            **b**: The right-hand-side vector for setting the linear inequalities.
        """
        if self.Y is None:
            self.Y = self.get_zonotope_vertices()
        n = self.Y.shape[1]
        if n == 1:
            A = np.array([[1],[-1]])
            b = np.array([[max(self.Y)],[min(self.Y)]])
            return  A, b
        else:
            convexHull = ConvexHull(self.Y)
            A = convexHull.equations[:,:n]
            b = -convexHull.equations[:,n]
            return A, b

    def get_samples_constraining_active_coordinates(self, inactive_samples, active_coordinates):
        """

        A hit and run type sampling strategy for generating samples at a given coordinate in the active subspace
        by varying its coordinates along the inactive subspace.

        :param Subspaces self:
            An instance of the Subspaces object.
        :param int inactive_samples:
            The number of inactive samples required.
        :param numpy.ndarray active_coordiantes:
            The active subspace coordinates.

        :return:
            **X**: An numpy.ndarray of the full-space coordinates.

        **Note:**
        This routine has been adapted from Paul Constantine's hit_and_run() function; see reference below.

        Constantine, P., Howard, R., Glaws, A., Grey, Z., Diaz, P., Fletcher, L., (2016) Python Active-Subspaces Utility Library. Journal of Open Source Software, 1(5), 79. `Paper <http://joss.theoj.org/papers/10.21105/joss.00079>`__.

        """
        y = active_coordinates
        N = inactive_samples
        W1 = self._subspace[:, :self.subspace_dimension]
        W2 = self._subspace[:, self.subspace_dimension:]
        m, n = W1.shape

        s = np.dot(W1, y).reshape((m, 1))
        normW2 = np.sqrt(np.sum(np.power(W2, 2), axis=1)).reshape((m, 1))
        A = np.hstack((np.vstack((W2, -W2.copy())), np.vstack((normW2, normW2.copy()))))
        b = np.vstack((1 - s, 1 + s)).reshape((2 * m, 1))
        c = np.zeros((m - n + 1, 1))
        c[-1] = -1.0
        # print()

        zc = linear_program_ineq(c, -A, -b)
        z0 = zc[:-1].reshape((m - n, 1))

        # define the polytope A >= b
        s = np.dot(W1, y).reshape((m, 1))
        A = np.vstack((W2, -W2))
        b = np.vstack((-1 - s, -1 + s)).reshape((2 * m, 1))

        # tolerance
        ztol = 1e-6
        eps0 = ztol / 4.0

        Z = np.zeros((N, m - n))
        for i in range(N):

            # random direction
            bad_dir = True
            count, maxcount = 0, 50
            while bad_dir:
                d = np.random.normal(size=(m - n, 1))
                bad_dir = np.any(np.dot(A, z0 + eps0 * d) <= b)
                count += 1
                if count >= maxcount:
                    Z[i:, :] = np.tile(z0, (1, N - i)).transpose()
                    yz = np.vstack([np.repeat(y[:, np.newaxis], N, axis=1), Z.T])
                    return np.dot(self._subspace, yz).T

            # find constraints that impose lower and upper bounds on eps
            f, g = b - np.dot(A, z0), np.dot(A, d)

            # find an upper bound on the step
            min_ind = np.logical_and(g <= 0, f < -np.sqrt(np.finfo(np.float).eps))
            eps_max = np.amin(f[min_ind] / g[min_ind])

            # find a lower bound on the step
            max_ind = np.logical_and(g > 0, f < -np.sqrt(np.finfo(np.float).eps))
            eps_min = np.amax(f[max_ind] / g[max_ind])

            # randomly sample eps
            eps1 = np.random.uniform(eps_min, eps_max)

            # take a step along d
            z1 = z0 + eps1 * d
            Z[i, :] = z1.reshape((m - n,))

            # update temp var
            z0 = z1.copy()

        yz = np.vstack([np.repeat(y[:, np.newaxis], N, axis=1), Z.T])
        return np.dot(self._subspace, yz).T

#    def plot_sufficient_summary(self, **kwargs):
    def plot_sufficient_summary(self, ax=None, X_test=None, y_test=None, show=True, poly=True, uncertainty=False, legend=False, scatter_kwargs={}, plot_kwargs={}):
        """
        Generates a sufficient summary plot for 1D or 2D polynomial ridge approximations.
        
        :param Subspace mysubspace:
            An instance of the Subspace class.
        :param matplotlib.ax ax: 
            An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
        :param :numpy.ndarray X_test:
            A numpy ndarray containing input test data to plot (in addition to the training data). Must have same number of dimensions as the training data. 
        :param :numpy.ndarray y_test:
            A numpy ndarray containing output test data to plot (in addition to the training data).
        :param bool show:
            Option to view the plot.
        :param bool poly:
            Option to plot the subspace polynomial.
        :param bool uncertainty:
            Option to show confidence intervals (1 standard deviation) of the subspace polynomial.
        :param bool legend:
            Option to show legend.
        :param dict scatter_kwargs:
            Dictionary of keyword arguments to pass to matplotlib.scatter functions.  
        :param dict plot_kwargs:
            Dictionary of keyword arguments to pass to matplotlib.plot functions.  
        """
        #return plot.plot_sufficient_summary(self,**kwargs)
        return plot.plot_sufficient_summary(self, ax, X_test, y_test, show, poly, uncertainty, legend, scatter_kwargs, plot_kwargs)

    def plot_2D_contour_zonotope(self, **kwargs):
        """
        Generates a 2D contour plot of the polynomial ridge approximation.
        """
        return plot.plot_2D_contour_zonotope(self,**kwargs)

    def plot_samples_from_second_subspace_over_first(self, mysubspace_2, **kwargs):
        """
        Generates a zonotope plot where samples from the second subspace are projected
        over the first.
        """
        return plot.plot_samples_from_second_subspace_over_first(self,mysubspace_2, **kwargs)

def vandermonde(eta, p):
    # TODO: Try using a "correlated" basis here?
    _, n = eta.shape
    listing = []
    for i in range(0, n):
        listing.append(p)
    Object=Basis('total-order',listing)
    # Establish n Parameter objects
    params = []
    P = Parameter(order=p, lower=-1, upper=1, distribution='uniform')
    for i in range(0, n):
        params.append(P)
    # Use the params list to establish the Poly object
    poly_obj = Poly(params, Object, method='least-squares')
    V = poly_obj.get_poly(eta)
    V = V.T
    return V, poly_obj


def vector_AS(list_of_polys, R = None, alpha=None, k=None, samples=None, bootstrap=False, bs_trials = 50
                , J = None, save_path = None):
    # Find AS directions to vector val func
    # analogous to computeActiveSubspace
    # Since we are dealing with *one* vector val func we should have just one input space
    # Take the first of the polys.
    poly = list_of_polys[0]
    if samples is None:
        d = poly.dimensions
        if alpha is None:
            alpha = 4
        if k is None or k > d:
            k = d
        M = int(alpha * k * np.log(d))
        X = np.zeros((M, d))
        for j in range(0, d):
            X[:, j] = np.reshape(poly.parameters[j].getSamples(M), M)
    else:
        X = samples
        M, d = X.shape
    n = len(list_of_polys) # number of outputs
    if R is None:
        R = np.eye(n)
    elif len(R.shape) == 1:
        R = np.diag(R)
    if J is None:
        J = jacobian_vec(list_of_polys,X)
        if not(save_path is None):
            np.save(save_path,J)


    J_new = np.matmul(sqrtm(R), np.transpose(J,[2,0,1]))
    JtJ = np.matmul(np.transpose(J_new,[0,2,1]), J_new)
    H = np.mean(JtJ,axis=0)

    # Compute P_r by solving generalized eigenvalue problem...
    # Assume sigma = identity for now
    e, W = np.linalg.eigh(H)
    eigs = np.flipud(e)
    eigVecs = np.fliplr(W)
    if bootstrap:
        all_bs_eigs = np.zeros((bs_trials, d))
        all_bs_W = []
        for t in range(bs_trials):
            print("Starting bootstrap trial %d"%t)
            bs_samples = X[np.random.randint(0,M,size=M), :]
            J_bs = jacobian_vec(list_of_polys, bs_samples)
            J_new_bs = np.matmul(sqrtm(R), np.transpose(J_bs,[2,0,1]))
            JtJ_bs = np.matmul(np.transpose(J_new_bs, [0, 2, 1]), J_new_bs)
            H_bs = np.mean(JtJ_bs, axis=0)

            # Compute P_r by solving generalized eigenvalue problem...
            # Assume sigma = identity for now
            e_bs, W_bs = np.linalg.eigh(H_bs)
            all_bs_eigs[t,:] = np.flipud(e_bs)
            eigVecs_bs = np.fliplr(W_bs)
            all_bs_W.append(eigVecs_bs)

        eigs_bs_lower = np.min(all_bs_eigs, axis = 0)
        eigs_bs_upper = np.max(all_bs_eigs, axis = 0)
        return eigs,eigVecs,eigs_bs_lower,eigs_bs_upper, all_bs_W
    else:
        return eigs,eigVecs


def jacobian_vp(V, V_plus, U, f, Polybasis, eta, minmax, X):
    M, N = V.shape
    m, n = U.shape
    Gradient = Polybasis.get_poly_grad(eta)
    sub = (minmax[1,:]-minmax[0,:]).T
    vectord = np.reshape(2.0/sub,(n,1))
    # Initialize the tensor
    J = np.zeros((M, m, n))
    # Obtain the derivative of this tensor
    dV = np.zeros((m, n, M, N))
    for l in range(0, n):
        for j in range(0, N):
            current = Gradient[l].T
            if n == 1:
                current = Gradient.T
            dV[:,l,:,j] = np.asscalar(vectord[l])*(X.T*current[:,j])

    # Get the P matrix
    P = np.identity(M)-np.matmul(V,V_plus)
    V_minus = scipy.linalg.pinv(V)

    # Calculate entries for the tensor
    for j in range(0,m):
        for k in range(0,n):
            temp1 = np.linalg.multi_dot([P,dV[j,k,:,:],V_minus])
            J[:, j, k]=(-np.matmul((temp1+temp1.T),f)).reshape((M,))  # Eqn 15

    return J


def jacobian_vec(list_of_poly, X):
    m = len(list_of_poly)
    [N, d] = X.shape
    J = np.zeros((m, d, N))
    for p in range(len(list_of_poly)):
        J[p, :, :] = list_of_poly[p].get_polyfit_grad(X)
    return J


def subspace_dist(U, V):
    if len(U.shape) == 1:
        return np.linalg.norm(np.outer(U, U) - np.outer(V, V), ord=2)
    else:
        return np.linalg.norm(np.dot(U, U.T) - np.dot(V, V.T), ord=2)


def linear_program_ineq(c, A, b):
    c = c.reshape((c.size,))
    b = b.reshape((b.size,))

    # make unbounded bounds
    bounds = []
    for i in range(c.size):
        bounds.append((None, None))

    A_ub, b_ub = -A, -b
    res = linprog(c, A_ub=A_ub, b_ub=b_ub, bounds=bounds, options={"disp": False}, method='simplex')
    if res.success:
        return res.x.reshape((c.size, 1))
    else:
        np.savez('bad_scipy_lp_ineq_{:010d}'.format(np.random.randint(int(1e9))),
                 c=c, A=A, b=b, res=res)
        raise Exception('Scipy did not solve the LP. Blame Scipy.')


def get_unique_rows(X0):
    X1 = X0.view(np.dtype((np.void, X0.dtype.itemsize * X0.shape[1])))
    return np.unique(X1).view(X0.dtype).reshape(-1, X0.shape[1])



import numpy as np
from copy import deepcopy
from equadratures.parameter import Parameter
from equadratures.poly import Poly
from equadratures.basis import Basis
import equadratures.plot as plot
from urllib.parse import quote

class PolyTree(object):
        """
        Definition of a polynomial tree object.

        :param str splitting_criterion:
                The type of splitting_criterion to use in the fit function. Options include ``model_aware`` which fits polynomials for each candidate split, ``model_agnostic`` which uses a standard deviation based model-agnostic split criterion [1], and ``loss_gradient`` which uses a gradient based splitting criterion similar to that in [2].
        :param int max_depth:
                The maximum depth which the tree will grow to.
        :param int min_samples_leaf:
                The minimum number of samples per leaf node.
        :param int order:
                The order of the generated orthogonal polynomials.
        :param str basis:
                The type of index set used for the basis. Options include: ``univariate``, ``total-order``, ``tensor-grid``, ``sparse-grid`` and ``hyperbolic-basis``
        :param str search:
                The method of search to be used. Options are ``grid`` or ``exhaustive``.
        :param int samples:
                The interval between splits if ``grid`` search is chosen.
        :param bool verbose:
                For debugging
        :param bool all_data:
                Save data at all nodes (instead of only leaf nodes).
        :param list split_dims:
                List of dimensions along which to make splits.

        **Sample constructor initialisations**::

                import numpy as np
                from equadratures import *

                tree = polytree.PolyTree()

                X = np.loadtxt('inputs.txt')
                y = np.loadtxt('outputs.txt')

                tree.fit(X,y)

        **References**
                1. Wang, Y., Witten, I. H., (1997) Inducing Model Trees for Continuous Classes. In Proc. of the 9th European Conf. on Machine Learning Poster Papers. 128-137. `Paper <https://researchcommons.waikato.ac.nz/handle/10289/1183>`__
                2. Broelemann, K., Kasneci, G., (2019) A Gradient-Based Split Criterion for Highly Accurate and Transparent Model Trees. In Int. Joint Conf. on Artificial Intelligence (IJCAI). 2030-2037. `Paper <https://www.ijcai.org/Proceedings/2019/0281.pdf>`__
                3. Chan, T. F., Golub, G. H., LeVeque, R. J., (1983) Algorithms for computing the sample variance: Analysis and recommendations. The American Statistician. 37(3): 242–247. `Paper <https://www.tandfonline.com/doi/abs/10.1080/00031305.1983.10483115>`__
        """
        def __init__(self, splitting_criterion='model_aware', max_depth=5, min_samples_leaf=None, order=1, basis='total-order', search='exhaustive', samples=50, verbose=False, poly_method="least-squares", poly_solver_args={},all_data=False,split_dims=None,k=0.05):
                self.splitting_criterion = splitting_criterion
                self.max_depth = max_depth
                self.min_samples_leaf = min_samples_leaf
                self.order = order
                self.basis = basis
                self.tree = None
                self.search = search
                self.samples = samples
                self.verbose = verbose
                self.cardinality = None
                self.poly_method = poly_method
                self.poly_solver_args = poly_solver_args
                self.actual_max_depth = 0
                self.all_data = all_data
                self.k = k
                if split_dims is not None:
                        split_dims = [split_dims] if not isinstance(split_dims, list) else split_dims
                        assert all(isinstance(dim, int) for dim in split_dims), "split_dims should be a list if ints"
                self.split_dims = split_dims

                assert max_depth >= 0, "max_depth must be >= 0"
                assert order > 0, "order must be a postive integer" 
                assert samples > 0, "samples must be a postive integer"
                assert k > 0, "k must be a positive number"

        def get_splits(self):
                """
                Returns the list of splits made

                :param PolyTree self:
                    An instance of the PolyTree class.
                :return:
                        **splits**: A list of Splits made in the format of a nested list: [[split, dimension], ...]
                """

                def _search_tree(node, splits):
                        if node["children"]["left"] != None:
                                if [node["threshold"], node["j_feature"]] not in splits:
                                        splits.append([node["threshold"], node["j_feature"]])
                                splits = _search_tree(node["children"]["left"], splits)

                        if node["children"]["right"] != None:
                                if [node["threshold"], node["j_feature"]] not in splits:
                                        splits.append([node["threshold"], node["j_feature"]])
                                splits = _search_tree(node["children"]["right"], splits)

                        return splits

                return _search_tree(self.tree, [])

        def _split_data(self, j_feature, threshold, X, y):
                idx_left = np.where(X[:, j_feature] <= threshold)[0]
                idx_right = np.delete(np.arange(0, len(X)), idx_left)
                assert len(idx_left) + len(idx_right) == len(X)
                return (X[idx_left], y[idx_left]), (X[idx_right], y[idx_right])

        def get_polys(self):
                """
                Returns the list of polynomials used in the tree

                :param PolyTree self:
                    An instance of the PolyTree class.
                :return:
                        **polys**: A list of Poly objects
                """

                def _search_tree(node, polys):
                        if node["children"]["left"] == None and node["children"]["right"] == None:
                                polys.append(node["poly"])

                        if node["children"]["left"] != None:
                                polys = _search_tree(node["children"]["left"], polys)

                        if node["children"]["right"] != None:
                                polys = _search_tree(node["children"]["right"], polys)

                        return polys

                return _search_tree(self.tree, [])

        def fit(self, X, y):
                """
                Fits the tree to the provided data

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param numpy.ndarray X:
                        Training input data
                :param numpy.ndarray y:
                        Training output data
                """

                def _build_tree():

                        global index_node_global

                        def _splitter(node):
                                # Extract data
                                X, y = node["data"]
                                depth = node["depth"]
                                N, d = X.shape

                                # Dimensions to split along
                                if self.split_dims is None:
                                    self.split_dims = range(d)

                                # Find feature splits that might improve loss
                                did_split = False
                                if self.splitting_criterion == "model_aware":
                                        loss_best = node["loss"]
                                elif self.splitting_criterion == "model_agnostic" or self.splitting_criterion=="loss_gradient":
                                        loss_best = np.inf
                                else:
                                        raise Exception("invalid splitting_criterion")
                                data_best = None
                                polys_best = None
                                j_feature_best = None
                                threshold_best = None

                                if self.verbose:
                                        polys_fit = 0

                                # Perform threshold split search only if node has not hit max depth
                                if (depth >= 0) and (depth < self.max_depth):
                                        if self.splitting_criterion != "loss_gradient":

                                                for j_feature in range(d):

                                                        last_threshold = np.inf

                                                        if self.search == 'exhaustive':
                                                                threshold_search = X[:, j_feature]
                                                        elif self.search == 'grid':
                                                                if self.samples > N:
                                                                        samples = N
                                                                else:
                                                                        samples = self.samples
                                                                threshold_search = np.linspace(np.min(X[:,j_feature]), np.max(X[:,j_feature]), num=samples)
                                                        else:
                                                                raise Exception('Incorrect search type! Must be \'exhaustive\' or \'grid\'')

                                                        # Perform threshold split search on j_feature
                                                        for threshold in np.unique(np.sort(threshold_search)):

                                                                # Split data based on threshold
                                                                (X_left, y_left), (X_right, y_right) = self._split_data(j_feature, threshold, X, y)
                                                                #print(j_feature, threshold, X_left, X_right)
                                                                N_left, N_right = len(X_left), len(X_right)

                                                                # Do not attempt to split if split conditions not satisfied
                                                                if not (N_left >= self.min_samples_leaf and N_right >= self.min_samples_leaf):
                                                                        continue

                                                                # Compute weight loss function
                                                                if self.splitting_criterion == "model_aware":
                                                                        loss_left, poly_left = _fit_poly(X_left, y_left)
                                                                        loss_right, poly_right = _fit_poly(X_right, y_right)

                                                                        loss_split = (N_left*loss_left + N_right*loss_right) / N

                                                                        if self.verbose: polys_fit += 2

                                                                elif self.splitting_criterion == "model_agnostic":
                                                                        loss_split = np.std(y) - (N_left*np.std(y_left) + N_right*np.std(y_right)) / N

                                                                # Update best parameters if loss is lower
                                                                if loss_split < loss_best:
                                                                        did_split = True
                                                                        loss_best = loss_split
                                                                        if self.splitting_criterion == "model_aware": polys_best = [poly_left, poly_right]
                                                                        data_best = [(X_left, y_left), (X_right, y_right)]
                                                                        j_feature_best = j_feature
                                                                        threshold_best = threshold

                                        # Gradient based splitting criterion from ref. [2]
                                        else:
                                                # Fit a single poly to parent node
                                                loss, poly = _fit_poly(X, y)

                                                # Now run the splitting algo using gradients from this poly
                                                did_split, j_feature_best, threshold_best = self._find_split_from_grad(poly, X, y.reshape(-1,1))

                                # If model_agnostic or gradient based, fit poly's to children now we have split
                                if self.splitting_criterion != "model_aware" and did_split:
                                        (X_left, y_left), (X_right, y_right) = self._split_data(j_feature_best, threshold_best, X, y)
                                        loss_left, poly_left = _fit_poly(X_left, y_left)
                                        loss_right, poly_right = _fit_poly(X_right, y_right)
                                        N_left, N_right = len(X_left), len(X_right)
                                        loss_best = (N_left*loss_left + N_right*loss_right) / N
                                        polys_best = [poly_left, poly_right]
                                        if self.splitting_criterion == "loss_gradient": data_best = [(X_left, y_left), (X_right, y_right)]

                                        if self.verbose: polys_fit += 2

                                if self.verbose and did_split: print("Node (X.shape = {}) fitted with {} polynomials generated".format(X.shape, polys_fit))
                                elif self.verbose: print("Node (X.shape = {}) failed to fit after {} polynomials generated".format(X.shape, polys_fit))

                                if did_split and depth > self.actual_max_depth:
                                        self.actual_max_depth = depth

                                # Return the best result
                                result = {"did_split": did_split,
                                                  "loss": loss_best,
                                                  "polys": polys_best,
                                                  "data": data_best,
                                                  "j_feature": j_feature_best,
                                                  "threshold": threshold_best,
                                                  "N": N}

                                return result

                        def _fit_poly(X, y):

#                                try:

                                N, d = X.shape
                                myParameters = []

                                for dimension in range(d):
                                        values = X[:,dimension]
                                        values_min = np.amin(values)
                                        values_max = np.amax(values)

                                        if (values_min - values_max) ** 2 < 0.01:
                                                myParameters.append(Parameter(distribution='Uniform', lower=values_min-0.01, upper=values_max+0.01, order=self.order))
                                        else:
                                                myParameters.append(Parameter(distribution='Uniform', lower=values_min, upper=values_max, order=self.order))
                                if self.basis == "hyperbolic-basis":
                                        myBasis = Basis(self.basis, orders=[self.order for _ in range(d)], q=0.5)
                                else:
                                        myBasis = Basis(self.basis, orders=[self.order for _ in range(d)])

                                container["index_node_global"] += 1
                                poly = Poly(myParameters, myBasis, method=self.poly_method, sampling_args={'sample-points':X, 'sample-outputs':y}, solver_args=self.poly_solver_args)
                                poly.set_model()

                                mse = np.linalg.norm(y - poly.get_polyfit(X).reshape(-1)) ** 2 / N
#                                except Exception as e:
#                                        print("Warning fitting of Poly failed:", e)
#                                        print(d, values_min, values_max)
#                                        mse, poly = np.inf, None

                                return mse, poly

                        def _create_node(X, y, depth, container):
                                poly_loss, poly = _fit_poly(X, y)

                                node = {"name": "node",
                                                "index": container["index_node_global"],
                                                "loss": poly_loss,
                                                "poly": poly,
                                                "data": (X, y),
                                                "n_samples": len(X),
                                                "j_feature": None,
                                                "threshold": None,
                                                "children": {"left": None, "right": None},
                                                "depth": depth,
                                                "flag": False}
                                container["index_node_global"] += 1

                                return node

                        def _split_traverse_node(node, container):

                                result = _splitter(node)
                                if not result["did_split"]:
                                        return

                                node["j_feature"] = result["j_feature"]
                                node["threshold"] = result["threshold"]

                                if not self.all_data:
                                    del node["data"]

                                (X_left, y_left), (X_right, y_right) = result["data"]
                                poly_left, poly_right = result["polys"]

                                node["children"]["left"] = _create_node(X_left, y_left, node["depth"]+1, container)
                                node["children"]["right"] = _create_node(X_right, y_right, node["depth"]+1, container)
                                node["children"]["left"]["poly"] = poly_left
                                node["children"]["right"]["poly"] = poly_right

                                # Split nodes
                                _split_traverse_node(node["children"]["left"], container)
                                _split_traverse_node(node["children"]["right"], container)

                        container = {"index_node_global": 0}
                        root = _create_node(X, y, 0, container)
                        _split_traverse_node(root, container)

                        return root

                N, d = X.shape
                if self.basis == "hyperbolic-basis":
                        self.cardinality = Basis(self.basis, orders=[self.order for _ in range(d)], q=0.5).get_cardinality()
                else:
                        self.cardinality = Basis(self.basis, orders=[self.order for _ in range(d)]).get_cardinality()
                if self.min_samples_leaf == None or self.min_samples_leaf == self.cardinality:
                        self.min_samples_leaf = int(np.ceil(self.cardinality * 1.25))
                elif self.cardinality > self.min_samples_leaf:
                        print("WARNING: Basis cardinality ({}) greater than the minimum samples per leaf ({}). This may cause reduced performance.".format(self.cardinality, self.min_samples_leaf))

                self.k *= self.min_samples_leaf

                self.tree = _build_tree()

        def prune(self, X, y, tol=0.0, percent=False):
                """
                Prunes the tree that you have fitted.

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param numpy.ndarray X:
                        Training input data
                :param numpy.ndarray y:
                        Training output data
                :param float tol:
                        Pruning tolerance (%). Prune nodes if they only improve loss by less than this tolerance (optional).
                :param bool percent:
                        If true, tol is taken as a percentage of the parent node's error. Otherwise, tol is taken to be an absolute value.
                """
                if percent: tol /= 100.0
                def pruner(node, X_subset, y_subset):

                        if X_subset.shape[0] < 1:
                                node["test_loss"] = 0
                                node["n_samples"] = 0
                                return node

                        node["test_loss"] = np.linalg.norm(y_subset - node["poly"].get_polyfit(X_subset).reshape(-1)) ** 2 / X_subset.shape[0]

                        is_left = node["children"]["left"] != None
                        is_right = node["children"]["right"] != None

                        if is_left and is_right:
                                (X_left, y_left), (X_right, y_right) = self._split_data(node["j_feature"], node["threshold"], X_subset, y_subset)

                                node["children"]["left"] = pruner(node["children"]["left"], X_left, y_left)
                                node["children"]["right"] = pruner(node["children"]["right"], X_right, y_right)

                                lower_loss = ( node["children"]["left"]["test_loss"] * node["children"]["left"]["n_samples"] + node["children"]["right"]["test_loss"] * node["children"]["right"]["n_samples"] ) / ( node["children"]["left"]["n_samples"] + node["children"]["right"]["n_samples"] )

                                node["lower_loss"] = lower_loss

                                if percent:
                                    loss_eps = tol * node["test_loss"]
                                else:
                                    loss_eps = tol
                                print(lower_loss + loss_eps, node["test_loss"])
                                if lower_loss + loss_eps > node["test_loss"]:
                                        if self.verbose: print("prune",lower_loss, node["test_loss"], node["children"]["left"]["test_loss"], node["children"]["left"]["n_samples"], node["children"]["right"]["test_loss"], node["children"]["right"]["n_samples"])
                                        node["children"]["left"] = None
                                        node["children"]["right"] = None

                        return node

                assert self.tree is not None, "Run fit() before prune()"
                (X_left, y_left), (X_right, y_right) = self._split_data(self.tree["j_feature"], self.tree["threshold"], X, y)

                self.tree["children"]["left"] = pruner(self.tree["children"]["left"], X_left, y_left)
                self.tree["children"]["right"] = pruner(self.tree["children"]["right"], X_right, y_right)


        def predict(self, X):
            """
            Evaluates the the polynomial tree approximation of the data.
            :param numpy.ndarray X:
                An ndarray with shape (number_of_observations, dimensions) at which the tree fit must be evaluated at.
            :return: **y**:
                A numpy.ndarray of shape (1, number_of_observations) corresponding to the polynomial approximation of the tree.
            """

            def _predict(node, indexes):

                y_pred[indexes, node["depth"], 0] = node["poly"].get_polyfit(X[indexes]).reshape(-1)
                y_pred[indexes, node["depth"], 1] = np.full(fill_value=node["n_samples"], shape=len(indexes))

                no_children = node["children"]["left"] is None and \
                              node["children"]["right"] is None
                if no_children: return

                idx_left = np.where(X[indexes, node["j_feature"]] <= node["threshold"])[0]
                idx_right = np.where(X[indexes, node["j_feature"]] > node["threshold"])[0]

                _predict(node["children"]["left"], indexes[idx_left])
                _predict(node["children"]["right"], indexes[idx_right])

            assert self.tree is not None
            y_pred = np.empty(shape=(X.shape[0], self.actual_max_depth + 2, 2)) * np.nan

            _predict(self.tree, np.arange(0, X.shape[0]))

            smoothed_y_pred = np.zeros(shape=(X.shape[0]))

            for y in range(0,X.shape[0]):
                i = self.actual_max_depth + 1

                while np.isnan(y_pred[y][i][0]) and i > 0:
                    i-=1

                smoothed_y = y_pred[y][i][0]

                #print(y_pred[i])
                while i > 0:
                    n_i = y_pred[y][i][1]
                    if n_i == 0: break
                    #print(smoothed_y)
                    smoothed_y = (smoothed_y * n_i + y_pred[y][i][0] * self.k) / (self.k + n_i)
                    i-=1

                #print("\n")
                smoothed_y_pred[y] = smoothed_y

            return smoothed_y_pred

        def apply(self,X):
                """
                Returns the leaf node index for each observation in the data.

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param numpy.ndarray X:
                    An ndarray with shape (number_of_observations, dimensions) at which the tree fit must be evaluated at.

                :return:
                **inode**: A numpy.ndarray of shape (number_of_observations,1) corresponding to the node indices for each observation in X.
                """
                def _apply(node, indexes):
                        no_children = node["children"]["left"] is None and \
                        node["children"]["right"] is None
                        if no_children:
                                inode[indexes] = node["index"]
                                return

                        idx_left = np.where(X[indexes, node["j_feature"]] <= node["threshold"])[0]
                        idx_right = np.where(X[indexes, node["j_feature"]] > node["threshold"])[0]
                        _apply(node["children"]["left"], indexes[idx_left])
                        _apply(node["children"]["right"], indexes[idx_right])

                if X.ndim == 1: X = X.reshape(1,-1)
                inode = np.zeros(shape=X.shape[0],dtype=int)
                _apply(self.tree, np.arange(0, X.shape[0]))
                return inode

        def get_leaves(self):
                """
                Returns the node indices for all leaf nodes.

                :param PolyTree self:
                    An instance of the PolyTree class.

                :return:
                **inode**: A list containing the node indices of all leaf nodes.
                """
                def _recurse(node,leaf_list):
                    no_children = node["children"]["left"] is None and \
                    node["children"]["right"] is None
                    if no_children:
                        leaf_list.append(node["index"])
                        return
                    _recurse(node["children"]["left"],leaf_list)
                    _recurse(node["children"]["right"],leaf_list)
                
                leaf_list = []
                _recurse(self.tree,leaf_list)
                return leaf_list

        def get_mean_and_variance(self):
            """
            Computes the mean and variance of the polynomial tree model.

            :param Poly self:
                An instance of the PolyTree class.

            :return:
                **mean**: The approximated mean of the polynomial tree fit; output as a float.

                **variance**: The approximated variance of the polynomial tree fit; output as a float.
            """
            # Get volume of polytree domain
            root_poly = self.tree["poly"]
            root_vol = self._calc_domain_vol(root_poly)

            # Get leaf nodes
            leaves = self.get_leaves()

            # Summation over all leaf nodes in the tree
            mean = 0.
            var  = 0.
            for leaf in leaves:
                leaf_poly = self.get_node(leaf)["poly"]
                leaf_vol = self._calc_domain_vol(leaf_poly)
                coeffs = leaf_poly.coefficients

                # Compute mean
                mean += (leaf_vol/root_vol) * float(coeffs[0])

                # Compute variance
                tmp = 0.
                for i in range(0, len(coeffs)):
                    tmp += float(coeffs[i]**2)
                var += (leaf_vol/root_vol) * tmp
            var -= mean**2 
        
            return mean, var

        def get_graphviz(self, X=None, feature_names=None, file_name=None):
                """
                Returns a url to the rendered graphviz representation of the tree.

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param numpy.ndarray X:
                        An ndarray with shape (dimensions) containing an input vector for a given sample, to highlight in the tree (optional).
                :param list feature_names:
                        A list of the names of the features used in the training data (optional).
                :param string filename:
                        Filename to write graphviz data to (optional). If None (default) then rendered in-place.
                """
                from graphviz import Digraph
                g = Digraph('g', node_attr={'shape': 'record', 'height': '.1'})

                if feature_names is None:
                    dim = self.tree["poly"].dimensions
                    feature_names = ['x_%d'%i for i in range(dim)]

                def _build_graphviz_recurse(node, parent_node_index=0, parent_depth=0, edge_label=""):

                        # Empty node
                        if node is None:
                                return

                        # Create node
                        node_index = node["index"]
                        if node["children"]["left"] is None and node["children"]["right"] is None:
                                threshold_str = ""
                                leaf = True
                        else:
                                threshold_str = "{} <= {:.3f}\\n".format(feature_names[node['j_feature']], node["threshold"])
                                leaf = False

                        if "lower_loss" in node:
                                label_str = "node {} \\n {} n_samples = {}\\n loss = {:.6f}\\n lower_loss = {}".format(node_index,threshold_str, node["n_samples"], node["test_loss"], node["lower_loss"])
                        elif "test_loss" in node:
                                label_str = "node {} \\n {} n_samples = {}\\n loss = {:.6f}".format(node_index,threshold_str, node["n_samples"], node["test_loss"])
                        else:
                                label_str = "node {} \\n {} n_samples = {}\\n loss = {:.6f}".format(node_index,threshold_str, node["n_samples"], node["loss"])
                        # Create node
                        if leaf:
                            nodeshape = "rectangle"
                            style     = ["rounded"]
                            fillcolor = "#E4fEE4"
                        else:
                            nodeshape = "rectangle"
                            style     = ["filled"]
                            fillcolor = "#EBFAFF"
                        if node["flag"]:
                            style.append('bold')
                        bordercolor = "black"
                        fontcolor = "black"
                        g.attr('node', label=label_str, shape=nodeshape)
                        g.node('node{}'.format(node_index),
                                   color=bordercolor, style=', '.join(style),
                                   fillcolor=fillcolor, fontcolor=fontcolor)

                        # Create edge
                        if parent_depth > 0:
                                if node["flag"]:
                                    edgecolor = 'orange'
                                    style     = 'bold'
                                else:
                                    edgecolor = 'black'
                                    style     = 'solid'
                                g.edge('node{}'.format(parent_node_index),
                                           'node{}'.format(node_index), label=edge_label, color=edgecolor,style=style)

                        # Traverse child or append leaf value
                        _build_graphviz_recurse(node["children"]["left"],
                                                                   parent_node_index=node_index,
                                                                   parent_depth=parent_depth + 1,
                                                                   edge_label="")
                        _build_graphviz_recurse(node["children"]["right"],
                                                                   parent_node_index=node_index,
                                                                   parent_depth=parent_depth + 1,
                                                                   edge_label="")

                def _flag_tree_walk(node,X):
                        node["flag"] = True
                        if node["children"]["left"] is None and \
                              node["children"]["right"] is None:
                                return
                        if X[node["j_feature"]] <= node["threshold"]:
                                return _flag_tree_walk(node["children"]["left"],X)
                        if X[node["j_feature"]] > node["threshold"]:
                                return _flag_tree_walk(node["children"]["right"],X)

                # Flag the node path to highlight later
                if X is not None:
                        _flag_tree_walk(self.tree,X)

                # Build graph
                _build_graphviz_recurse(self.tree,
                                                           parent_node_index=0,
                                                           parent_depth=0,
                                                           edge_label="")

                if file_name is None:
                        try:
                                g.render(view=True)
                        except:
                                file_name = 'tree.dot'
                                print("GraphViz source file written to " + file_name + " and can be viewed using an online renderer. Alternatively you can install graphviz on your system to render locally")

                if file_name is not None: # not elif here as file_name might be updated in try-except above
                        with open(file_name, "w") as file:
                                file.write(str(g.source))

        def get_node(self, inode):
                """
                Returns the node corresponding to a given node number inode (int).

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param int inode:
                        An int containing the node index.
                :return:
                **node**: The data for the node X belongs to; output as a dict.
                """
                # Find node with given index inode. Traverse all children until correct node found.
                def _get_node_from_n(node):
                        if node is not None: # Need to check if node is None here as below _get_node_from_n() calls on children will result in None if leaf node
                                if node["index"] == inode:
                                        return node
                                else:
                                        result = _get_node_from_n(node["children"]["right"])
                                        if result is None:
                                                result = _get_node_from_n(node["children"]["left"])
                                        return result
                        else:
                                return None
                return _get_node_from_n(self.tree)

        def get_paths(self,X=None):
                """
                Returns the tree paths for the leaf nodes in the tree.

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param numpy.ndarray X:
                    An ndarray with shape (number_of_observations, dimensions) to apply the tree to (optional). If given, paths will only be returned for leaves which contain observations.

                :return:
                **paths**: A dict containing a dict for each leaf node. Indexed by the node indexes for the leaf nodes.
                """

                def _find_path(node, path, i):
                        """
                        Private recursive function to find path through a tree for a given leaf node.
                        """
                        node_index = node["index"]
                        info = {'node':node_index,'j':node["j_feature"],'thresh':node["threshold"]}
                        path.append(info)
                        if node_index == i:
                                return True
                        left = False
                        right = False
                        if node["children"]["left"] is not None:
                                left = _find_path(node["children"]["left"], path, i)
                        if node["children"]["right"] is not None:
                                right = _find_path(node["children"]["right"], path, i)
                        if left or right :
                                return True
                        path.remove(info)
                        return False

                # Get leaf node id's
                if X is None:
                    leave_id = self.get_leaves()
                else:
                    # Get leaf nodes
                    leave_id = self.apply(X)

                # Loop through leaves and find path for each.
                paths ={}
                for leaf in np.unique(leave_id):
                        path_leaf = []
                        _find_path(self.tree, path_leaf, leaf)

                        # Set split info to None for leaf node
                        path_leaf[-1]["j"]      = None
                        path_leaf[-1]["thresh"] = None

                        # Save in dict
                        paths[leaf] = path_leaf

                return paths

        def plot_decision_surface(self,ij,ax=None,X=None,y=None,max_depth=None,label=True,
                                         color='data',colorbar=True,show=True,kwargs={}):
                """
                Plots the decision boundaries of the PolyTree over a 2D surface.
        
                :param PolyTree self: 
                    An instance of the PolyTree class.
                :param list ij: 
                    A list containing the two dimensions to plot over. For example, ``ij=[6,7]`` with plot over the 6th and 7th dimensions in ``X``.
                :param matplotlib.ax ax: 
                    An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
                :param :numpy.ndarray X:
                    A numpy ndarray containing the input data to plot.
                :param :numpy.ndarray y:
                    A numpy ndarray containing the output data to plot.
                :param int max_depth:
                    The maximum tree depth to plot decision boundaries for.
                :param bool label:
                    If ``True`` then the subdomains are labelled by their node number.
                :param string color:
                    What to color the scatter points by. ``'data'`` to color by the ``X``,``y`` data. ``'predict'`` to color by the PolyTree predictions, and ``'error'`` to color by the predictive error. (default: ``'data'``).
                :param bool colorbar:
                    Option to add a colorbar.
                :param bool show:
                    Option to view the plot.
                :param dict kwargs:
                    Dictionary of keyword arguments to pass to matplotlib.scatter().  
                """
                return plot.plot_decision_surface(self,ij,ax,X,y,max_depth,label,color,colorbar,show,kwargs)

        def _find_split_from_grad(self,model, X, y):
                """
                Private method to find the optimal split point for a tree node based on the training data in that node.

                :param PolyTree self:
                    An instance of the PolyTree class.
                :param Poly model:
                    An instance of the Poly class, corresponding to the Poly belonging to the tree node.
                :param numpy.ndarray X:
                        An ndarray with shape (number_of_observations, dimensions) containing the input data belonging to the tree node.
                :param numpy.ndarray y:
                        An ndarray with shape (number_of_observations, 1) containing the response data belonging to the tree node.
                :return:
                **did_split**: True if a split was found, otherwise False; output as a bool.
                **split_dim**: The dimension in X within which the best split was found; output as an int.
                **split_val**: The location of the best split; output as a float.
                """
                renorm = True
                N,D = np.shape(X)

                # Gradient of loss wrt model coefficients
                P = model.get_poly(X).T
                r = y-model.get_polyfit(X)
                g = r*P

                # Sum of gradients
                gsum = g.sum(axis=0)

                # Loop through all dimensions in X
                split_dim = None
                split_val = None
                gain_max  = -np.inf
                for d in self.split_dims:
                    # Sort along feature i
                    sort = np.argsort(X[:,d])
                    Xd   = X[sort,d]

                    # Find unique values along one column. #TODO - grid search option
                    _,splits = np.unique(Xd,return_index=True)
                    splits = splits[1:]

                    # Number of samples on left and right split
                    N_l = splits
                    N_r = N - N_l

                    # Only take splits where both children have more than `min_samples_leaf` samples
                    idx = np.minimum(N_l, N_r) >= self.min_samples_leaf
                    splits = splits[idx]
                    N_l    = N_l[idx].reshape(-1,1)
                    N_r    = N_r[idx].reshape(-1,1)

                    # If we've run out of candidate spilts, skip
                    if len(splits) <= 1:
                        continue

                    # Sums of gradients for left and right
                    gsum_left  = g[sort,:].cumsum(axis=0)
                    gsum_left  = gsum_left[splits-1,:]
                    gsum_right = gsum - gsum_left

                    # Renorm. gradients to zero mean and unit std
                    if renorm:
                        mu_l, mu_r, sigma_l, sigma_r = self._get_mean_and_sigma(P[:,1:],splits,N_l,N_r,sort)
                        gsum_left  = self._renormalise( gsum_left, 1/sigma_l, -mu_l/sigma_l)
                        gsum_right = self._renormalise(gsum_right, 1/sigma_r, -mu_r/sigma_r)

                    # Compute the Gain (see Eq. (6) in [1])
                    gain = (gsum_left**2).sum(axis=1)/N_l.reshape(-1) + (gsum_right**2).sum(axis=1)/N_r.reshape(-1)

                    # Find best gain and compare with previous best
                    best_idx = np.argmax(gain)
                    gain     = gain[best_idx]
                    if gain > gain_max:
                        gain_max  = gain
                        best_split_dim = d
                        best_split_val = 0.5*(Xd[splits[best_idx] - 1] + Xd[splits[best_idx]])

                # If gain_max stilll == -np.inf, we must have passed through all features w/o finding a split
                # so return False. Otherwise return True and the spilt details.
                if gain_max == -np.inf:
                    return False, None, None
                else:
                    return True, best_split_dim, best_split_val

        @staticmethod
        def _get_mean_and_sigma(X,splits,N_l,N_r,sort):
                """
                Computes mean and standard deviation of the data in array X, when it is
                split in two by the threshold values in the splits array. The data is offset by
                its mean to avoid catastrophic cancellation when computing the variance (see ref. [3]).
                X - [N,ndim] array of data.
                splits  - [Nsplit] array of split locations.
                sort   - [N] array reordering X.
                """
                # Min value of sigma (for stability later)
                epsilon = 0.001
        
                # Reorder, and shift X by mean
                mu     = np.reshape(np.mean(X, axis=0), (1, -1))
                Xshift = X[sort] - mu
        
                # Cumulative sums (and sums of squares) for left and right splits
                Xsum_l  = Xshift.cumsum(axis=0)
                Xsum_r  = Xsum_l[-1:,:] - Xsum_l
                X2sum_l = (Xshift**2).cumsum(axis=0)
                X2sum_r = X2sum_l[-1:,:] - X2sum_l
        
                # Compute mean of left and right side for all splits
                mu_l = Xsum_l[splits-1,:] / N_l
                mu_r = Xsum_r[splits-1,:] / N_r
        
                # Compute standard deviation of left and right side for all splits
                sigma_l = np.sqrt(np.maximum(X2sum_l[splits-1,:]/(N_l-1)-mu_l**2, epsilon**2))
                sigma_r = np.sqrt(np.maximum(X2sum_r[splits-1,:]/(N_r-1)-mu_r**2, epsilon**2))
        
                # Correct for previous shift
                mu_l = mu_l + mu
                mu_r = mu_r + mu
        
                return mu_l, mu_r, sigma_l, sigma_r
        
        @staticmethod
        def _renormalise(gradients, a, c):
                """
                Renormalises gradients according to according to eq. (14) of [1].
                Inputs
                ------
                gradients: array [n_samples, n_params] of gradients
                a: array [n_samples, n_params-1]: The normalisation factor
                c: array [n_samples, n_params-1]: The normalisation offset
                Returns
                -------
                gradients: array [n_samples, n_params]: Renormalised gradients
                """
                c = c*gradients[:,0].reshape(-1,1)
                gradients[:,1:] = gradients[:,1:] * a + c
                return gradients

        @staticmethod
        def _calc_domain_vol(Polynomial):
            params = Polynomial.parameters
            vol = 1.
            for param in params:
                vol *= param.upper - param.lower
            return vol




"""Plotting utilities."""
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl
from scipy.spatial import ConvexHull, convex_hull_plot_2d
from equadratures.datasets import score
import numpy as np
# seaborn defaults
sns.set(font_scale=1.5)
sns.set_style("white")
sns.set_style("ticks")
palette = sns.color_palette('deep')

# matplotlib defaults
mpl.rcParams['figure.facecolor'] = 'none'

def plot_sufficient_summary(mysubspace, ax=None, X_test=None, y_test=None, show=True, poly=True, uncertainty=False, legend=False, scatter_kwargs={}, plot_kwargs={}):
    """
    Generates a sufficient summary plot for 1D or 2D polynomial ridge approximations.
    
    :param Subspace mysubspace:
        An instance of the Subspace class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param :numpy.ndarray X_test:
        A numpy ndarray containing input test data to plot (in addition to the training data). Must have same number of dimensions as the training data. 
    :param :numpy.ndarray y_test:
        A numpy ndarray containing output test data to plot (in addition to the training data).
    :param bool show:
        Option to view the plot.
    :param bool poly:
        Option to plot the subspace polynomial.
    :param bool uncertainty:
        Option to show confidence intervals (1 standard deviation) of the subspace polynomial.
    :param bool legend:
        Option to show legend.
    :param dict scatter_kwargs:
        Dictionary of keyword arguments to pass to matplotlib.scatter().  
    :param dict plot_kwargs:
        Dictionary of keyword arguments to pass to matplotlib.plot().  
    """
    # Set default kwargs
    scatter_kwargs = set_defaults(scatter_kwargs,{'marker':'o','s':60,'ec':'k','lw':2,'alpha':0.7})
    plot_kwargs    = set_defaults(plot_kwargs   ,{'lw':3})

    # 1D or 2D ridge
    subdim = mysubspace.subspace_dimension

    if ax is None:
        if subdim == 1:
            fig,ax = plt.subplots(figsize=(7, 5),tight_layout=True)
            ax.set_xlabel(r'$u$')
            ax.set_ylabel(r'$y$')
        elif subdim == 2:
            fig = plt.figure(figsize=(9,7),tight_layout=True)
            ax = fig.add_subplot(111, projection='3d')
            ax.zaxis.set_rotate_label(False)
            ax.set_xlabel(r'$u_1$',labelpad=10)
            ax.set_ylabel(r'$u_2$',labelpad=10)
            ax.set_zlabel(r'$y$'  ,labelpad=10)
        else:
            raise ValueError("Currently can only generate sufficient summary plot for 1D and 2D ridge approximations")

    X = mysubspace.sample_points
    y = mysubspace.sample_outputs
    M = mysubspace.get_subspace()
    W = M[:,:subdim]
    u = X @ W
    if (X_test is not None) and (y_test is not None):
        u_test = X_test @ W
        test = True
    else:
        test = False
    if poly:
        subpoly = mysubspace.get_subspace_polynomial()

    if subdim == 1:
        ax.scatter(u, y, color=palette[0], label='Training samples', **scatter_kwargs)
        if test:
            ax.scatter(u_test, y_test, color=palette[2], label='Test samples', **scatter_kwargs)
        if poly:
            u_samples = np.linspace(np.min(u[:,0]), np.max(u[:,0]), 50)
            if uncertainty:
                y_mean, y_std = subpoly.get_polyfit(u_samples,uq=True)
                ax.fill_between(u_samples,(y_mean-y_std).squeeze(),(y_mean+y_std).squeeze(), color=palette[3], alpha=0.3, label=r'Polynomial  $\pm\sigma$')
            else:
                y_mean = subpoly.get_polyfit(u_samples)
            ax.plot(u_samples,y_mean,color=palette[3],label='Polynomial approx.',**plot_kwargs)

    elif subdim == 2:
        ax.scatter(u[:,0], u[:,1], y, color=palette[0], label='Training samples',**scatter_kwargs)
        if test:
            ax.scatter(u_test[:,0], u_test[:,1], y_test, color=palette[2], label='Test samples')
        if poly:
            subpoly = mysubspace.get_subspace_polynomial()
            N = 40
            u1_samples = np.linspace(np.min(u[:,0]), np.max(u[:,0]), N)
            u2_samples = np.linspace(np.min(u[:,1]), np.max(u[:,1]), N)
            [U1, U2] = np.meshgrid(u1_samples, u2_samples)
            u_samples = np.hstack([U1.reshape(N*N,1), U2.reshape(N*N,1)])
            if uncertainty:
                y_mean, y_std = subpoly.get_polyfit(u_samples, uq=True)
                y_mean = y_mean.reshape(N,N)
                y_std  = y_std.reshape(N,N)
                ax.plot_surface(U1, U2, y_mean-y_std, rstride=2, cstride=2, color=palette[3], ec='None', alpha=0.2, label='_nolegend_')
                ax.plot_surface(U1, U2, y_mean+y_std, rstride=2, cstride=2, color=palette[3], ec='None', alpha=0.2, label='_nolegend_')
            else:
                y_mean = subpoly.get_polyfit(u_samples).reshape(N,N)
            ax.plot_surface(U1, U2, y_mean, rstride=2, cstride=2, color=palette[3], ec='None', alpha=0.7, label='_nolegend_')

    sns.despine(offset=10, trim=True)

    if legend:
        ax.legend(ncol=2)
    if show:
        plt.show()
    if 'fig' in locals():
        return fig, ax
    else:
        return ax


def plot_2D_contour_zonotope(mysubspace, minmax=[-3.5, 3.5], grid_pts=180,  \
                             show=True, ax=None):
    """
    Generates a 2D contour plot of the polynomial ridge approximation.
    
    :param Subspace mysubspace:
        An instance of the Subspace class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param list minmax:
        A list of the minimum and maximum values of :math:`M^T x`, where :math:`M` is the subspace.
    :param int grid_pts:
        The number of grid points for generating the contour plot.
    :param bool show: 
        Option to view the plot.
    """
    if ax is None:
        fig,ax = plt.subplots(figsize=(10, 7),tight_layout=True)
        ax.set_xlabel(r'$u_1$')
        ax.set_ylabel(r'$u_2$')
    else:
        fig = ax.figure
   
    # Utilities for contour plot.
    x1 = np.linspace(minmax[0], minmax[1], grid_pts)
    XX, YY = np.meshgrid(x1, x1)
    xx1 = XX.reshape(grid_pts*grid_pts, )
    yy1 = YY.reshape(grid_pts*grid_pts, )
    H = np.vstack([xx1,yy1]).T
    
    A, b = mysubspace.get_linear_inequalities()
    subspacepoly = mysubspace.get_subspace_polynomial()
    H_evals = subspacepoly.get_polyfit(H)
    
    subspace_poly_evals = np.zeros((grid_pts*grid_pts, ))
    for i in range(0, grid_pts*grid_pts):
        u_coord = np.array([H[i, 0], H[i, 1]]).reshape(2, 1)
        b_out = A @ u_coord
        if np.all(b_out.flatten() <= b):
            subspace_poly_evals[i] = H_evals[i]
        else:
            subspace_poly_evals[i] = np.nan
    
    pts = mysubspace.get_zonotope_vertices()
    hull = ConvexHull(pts)
    u = mysubspace.sample_points @ mysubspace._subspace[:,0:2]
    
    # Contour plot!
    norm = mpl.colors.Normalize(vmin=np.min(mysubspace.sample_outputs) - 0.4 * np.std(mysubspace.sample_outputs),\
                                               vmax=np.max(mysubspace.sample_outputs)+ 0.4 * np.std(mysubspace.sample_outputs))
    ax.contourf(XX, YY, subspace_poly_evals.reshape(grid_pts,grid_pts), levels=40, linewidth=0, \
                antialiased=False, norm=norm)  
    c = ax.scatter(u[:,0],u[:,1], c=mysubspace.sample_outputs, marker='o', edgecolor='w', lw=1, alpha=0.8, s=80, norm=norm )  
    fig.colorbar(c, orientation="vertical", pad=0.1, shrink=0.5,label='y')
    for simplex in hull.simplices:
        ax.plot(pts[simplex, 0], pts[simplex, 1], 'k-', lw=5)
    sns.despine(offset=10, trim=True)
    if show:
        plt.show()
    return fig, ax
    
def plot_samples_from_second_subspace_over_first(mysubspace_1, mysubspace_2, axs=None, no_of_samples=500,\
                             minmax=[-3.5, 3.5], grid_pts=180, show=True): 
    """
    Generates a zonotope plot where samples from the second subspace are projected
    over the first.
    
    
    :param Subspace mysubspace_1:
        An instance of the Subspace class.
    :param Subspace mysubspace_2:
        A second instance of the Subspace class.
    :param list axs: 
        A len(2) list containing two instances of the ``matplotlib`` axes class to plot onto. If ``None``, new figures and axes are created (default: ``None``).
    :param int no_of_samples:
        Number of inactive samples to be generated.
    :param list minmax:
        A list of the minimum and maximum values of :math:`M^T x`, where :math:`M` is the subspace.
    :param int grid_pts:
        The number of grid points for generating the contour plot.
    :param bool show: 
        Option to view the plot.
    """    
    # 1. Generate a grid on the first zonotope.
    x1 = np.linspace(minmax[0], minmax[1], grid_pts)
    XX, YY = np.meshgrid(x1, x1)
    xx1 = XX.reshape(grid_pts*grid_pts, )
    yy1 = YY.reshape(grid_pts*grid_pts, )
    H = np.vstack([xx1,yy1]).T

    A, b = mysubspace_1.get_linear_inequalities()
    subspacepoly_1 = mysubspace_1.get_subspace_polynomial()
    subspacepoly_2 = mysubspace_2.get_subspace_polynomial()
    H_evals = subspacepoly_1.get_polyfit(H)

    subspace_poly_evals = np.zeros((grid_pts*grid_pts, ))
    for i in range(0, grid_pts*grid_pts):
        u_coord = np.array([H[i, 0], H[i, 1]]).reshape(2, 1)
        b_out = A @ u_coord
        if np.all(b_out.flatten() <= b):
            subspace_poly_evals[i] = H_evals[i]
        else:
            subspace_poly_evals[i] = np.nan

    indices = np.argwhere(~np.isnan(subspace_poly_evals))
    pts_inside = H[indices.flatten(),:]
    F = pts_inside.shape[0]

    # 2. For each point on this grid, generate no_of_sample points along the
    # inactive subspace
    mean_values = np.zeros((F, 1))
    std_values = np.zeros((F, 1))
    for counter in range(0, F):
        X_samples = mysubspace_1.get_samples_constraining_active_coordinates(no_of_samples, \
                                            pts_inside[counter, :].reshape(2,))
        y_values_for_samples = subspacepoly_2.get_polyfit(X_samples)
        mean_values[counter] = np.mean(y_values_for_samples)
        std_values[counter] = np.std(y_values_for_samples)

    pts_zono = mysubspace_1.get_zonotope_vertices()
    hull = ConvexHull(pts_zono)

    if axs is None:
        fig1,ax1 = plt.subplots(figsize=(10, 7),tight_layout=True)
        fig2,ax2 = plt.subplots(figsize=(10, 7),tight_layout=True)
        ax1.set_xlabel(r'$u_1$')
        ax1.set_ylabel(r'$u_2$')
        ax2.set_xlabel(r'$u_1$')
        ax2.set_ylabel(r'$u_2$')
    else:
        ax1 = axs[0]
        ax2 = axs[1]
        fig1 = ax1.figure
        fig2 = ax2.figure

    # Contour plot of mean values
    subspace_poly_evals[indices] =  mean_values
    c = ax1.contourf(XX, YY, subspace_poly_evals.reshape(grid_pts,grid_pts), levels=40, linewidth=0, \
                antialiased=False)  
    fig1.colorbar(c, orientation="vertical", pad=0.1, shrink=0.5,label=r'$\mu$')
    for simplex in hull.simplices:
        ax1.plot(pts_zono[simplex, 0], pts_zono[simplex, 1], 'k-', lw=5)

    # Contour plot of std values
    subspace_poly_evals[indices] =  std_values
    c = ax2.contourf(XX, YY, subspace_poly_evals.reshape(grid_pts,grid_pts), levels=40, linewidth=0, \
                antialiased=False)  
    fig2.colorbar(c, orientation="vertical", pad=0.1, shrink=0.5,label=r'$\sigma$')
    for simplex in hull.simplices:
        ax2.plot(pts_zono[simplex, 0], pts_zono[simplex, 1], 'k-', lw=5)

    sns.despine(fig1,offset=10, trim=True)
    sns.despine(fig2,offset=10, trim=True)
    if show:
        plt.show()
    return fig1, ax1, fig2, ax2

def plot_sobol(Polynomial, ax=None, order=1, show=True, labels=None, kwargs={}):
    """
    Plots a polynomial's Sobol' indices of a given order.

    :param Poly Polynomial: 
        An instance of the Poly class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param int order:
        Order of the Sobol' indices to plot.
    :param list parameters: 
        List of the parameters for the given polynomial.
    :param bool show: 
        Option to show the graph.
    :param dict kwargs:
        Dictionary of keyword arguments to pass to matplotlib.bar().  
    """
    # Set default kwargs
    kwargs = set_defaults(kwargs, {'color':'dodgerblue', 'ec':'k','lw':2, 'alpha':0.7})

    if ax is None:
        fig,ax = plt.subplots(figsize=(9, 6),tight_layout=True)
    else:
        fig = ax.figure

    ndims = Polynomial.dimensions
    sobol_indices=Polynomial.get_sobol_indices(order)
    nsob = len(sobol_indices)
    if (order > 3):
        raise ValueError("Only Sobol' indices of order<=3 can currently be plotted")
    if (nsob == 0):
        raise ValueError("Insufficient number of parameters to obtain order=%d Sobol' indices" %order)

    if order == 1:
        ax.set_ylabel(r'$S_i$')
        if labels is None: labels=[r'$S_%d$' %i for i in range(ndims)]
        to_plot = [sobol_indices[(i,)] for i in range(ndims)]

    elif order == 2:
        ax.set_ylabel(r'$S_{ij}$')
        if labels is None:
            labels=[r'$S_{%d%d}$' %(i,j) for i in range(ndims) for j in range(i+1,ndims)]
        else:
            labels=[labels[i] + r' $\leftrightarrow$ ' + labels[j]  for i in range(ndims) for j in range(i+1,ndims)]
        to_plot = [sobol_indices[(i,j)] for i in range(ndims) for j in range(i+1,ndims)]

    elif order == 3:
        ax.set_ylabel(r'$S_{ijk}$')
        if labels is None:
            labels=[r'$S_{%d%d%d}$' %(i,j,k) for i in range(ndims) for j in range(i+1,ndims) for k in range(j+1,ndims)]
        else:
            labels=[labels[i] + r' $\leftrightarrow$ ' + labels[j] + r' $\leftrightarrow$ ' + labels[k]
                    for i in range(ndims) for j in range(i+1,ndims) for k in range(j+1,ndims)]
        to_plot = [sobol_indices[(i,j,k)] for i in range(ndims) for j in range(i+1,ndims) for k in range(j+1,ndims)]

    ax.set_xticks(np.arange(nsob))
    ax.set_xticklabels(labels,rotation=45,rotation_mode="anchor",ha='right')
    ax.bar(np.arange(nsob),to_plot,**kwargs)
    sns.despine(fig, offset=10, trim=True)

    if show:
        plt.show()
    return fig, ax

def plot_total_sobol(Polynomial, ax=None, show=True, labels=None, kwargs={}):
    """
    Plots a polynomial's total-order Sobol' indices.

    :param Poly Polynomial: 
        An instance of the Poly class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param list parameters: 
        List of the parameters for the given polynomial.
    :param bool show: 
        Option to show the graph.
    :param dict kwargs:
        Dictionary of keyword arguments to pass to matplotlib.bar().  
    """
    # Set default kwargs
    kwargs = set_defaults(kwargs, {'color':'dodgerblue', 'ec':'k','lw':2, 'alpha':0.7})

    if ax is None:
        fig,ax = plt.subplots(figsize=(9, 6),tight_layout=True)

    ndims = Polynomial.dimensions
    sobol_indices=Polynomial.get_total_sobol_indices()

    ax.set_ylabel(r'$S_{T_i}$')
    if labels is None: labels=[r'$S_{T_%d}$' %i for i in range(ndims)]
    to_plot = [sobol_indices[(i,)] for i in range(ndims)]

    ax.set_xticks(np.arange(ndims))
    ax.set_xticklabels(labels,rotation=45,rotation_mode="anchor",ha='right')
    ax.bar(np.arange(ndims),to_plot,**kwargs)
    sns.despine(fig, offset=10, trim=True)

    if show:
        plt.show()
    if 'fig' in locals():
        return fig, ax
    else:
        return ax


def plot_pdf(Parameter, ax=None, data=None, show=True):
    """
    Plots the probability density function for a Parameter.

    :param Parameter Parameter: 
        An instance of the Parameter class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param numpy.array data: 
        Samples from the distribution (or a similar one) that need to be plotted as a histogram.
    """
    if ax is None:
        fig,ax = plt.subplots(figsize=(8, 6),tight_layout=True)
        ax.set_xlabel(Parameter.variable.capitalize())
        ax.set_ylabel('PDF')
    else:
        fig = ax.figure
    s_values, pdf = Parameter.get_pdf()
    ax.fill_between(s_values,  pdf*0.0, pdf, color="gold" , label='Density', interpolate=True, hatch="\\\\\\\\", edgecolor="grey",  linewidth=0.5,alpha=0.5)
    if data is not None:
        ax.hist(data, 50, density=True, facecolor='dodgerblue', alpha=0.7, label='Data', edgecolor='white')
    ax.legend()
    sns.despine(offset=10, trim=True)
    if show:
        plt.show()
    return fig, ax

def plot_orthogonal_polynomials(Parameter, ax=None, order_limit=None, number_of_points=200, show=True):
    """
    Plots the first few orthogonal polynomials.

    :param Parameter Parameter: 
        An instance of the Parameter class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param int order_limit:
        The maximum number of orthogonal polynomials that need to be plotted.
    :param int number_of_points: 
        The number of points used for plotting.
    :param bool show: 
        Option to view the plot.

    **Example**::

        import numpy as np
        from equadratures import *

        myparam = eq.Parameter(distribution='uniform', lower = -1.0, upper = 1.0, order=8, endpoints='both')
        myparam.plot_orthogonal_polynomials()
        
    """
    Xi = np.linspace(Parameter.distribution.x_range_for_pdf[0], \
                Parameter.distribution.x_range_for_pdf[-1], number_of_points).reshape(number_of_points, 1)
    P, _, _ = Parameter._get_orthogonal_polynomial(Xi)
    if ax is None:
        fig,ax = plt.subplots(figsize=(8, 6),tight_layout=True)
        ax.set_xlabel(Parameter.name.capitalize()+' parameter')
        ax.set_ylabel('Orthogonal polynomials')
    else:
        fig = ax.figure
    if order_limit is None:
        max_order = P.shape[0]
    else:
        max_order = order_limit
    for i in range(0, max_order):
        ax.plot(Xi, P[i,:], '-', lw=2, label='order %d'%(i))
    ax.legend()
    sns.despine(offset=10, trim=True)
    if show:
        plt.show()
    return fig, ax

def plot_polyfit_1D(Polynomial, ax=None, uncertainty=True, output_variances=None, number_of_points=200, show=True):
    """
    Plots a 1D only polynomial fit to the data.

    :param Poly Polynomial: 
        An instance of the Polynomial class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param bool uncertainty:
        Option to show confidence intervals (1 standard deviation).
    :param numpy.array output_variances:
        User-defined uncertainty associated with each data point; can be either a ``float`` in which case all data points are assumed to have the same variance, or can be an array of length equivalent to the number of data points.
    :param bool show: 
        Option to view the plot.
    """
    if ax is None:
        fig,ax = plt.subplots(figsize=(8, 6),tight_layout=True)
        ax.set_xlabel(Polynomial.parameters[0].variable.capitalize())
        ax.set_ylabel('Polynomial fit')
    else:
        fig = ax.figure
    if Polynomial.dimensions != 1:
        raise(ValueError, 'plot_polyfit_1D is only meant for univariate polynomials.')
    Xi = np.linspace(Polynomial.parameters[0].distribution.x_range_for_pdf[0], \
                Polynomial.parameters[0].distribution.x_range_for_pdf[-1], number_of_points).reshape(number_of_points, 1)
    if uncertainty:
        if output_variances is None:
            y, ystd = Polynomial.get_polyfit(Xi,uq=True)
        else:
            Polynomial.output_variances = output_variances
            y, ystd = Polynomial.get_polyfit(Xi,uq=True)
        ystd = ystd.squeeze()
    else:
        y = Polynomial.get_polyfit(Xi)
    y = y.squeeze()
    X = Polynomial.get_points()
    y_truth = Polynomial._model_evaluations
    ax.plot(Xi, y, '-',label='Polynomial fit', color='navy')
    ax.plot(X.flatten(), y_truth.flatten(), 'o', color='dodgerblue', ms=10, markeredgecolor='k',lw=1, alpha=0.6, label='Data')
    if uncertainty:
        ax.fill_between(Xi.flatten(), y+ystd, y-ystd, alpha=.10, color='deepskyblue',label='Polynomial $\sigma$')
    ax.legend()
    sns.despine(offset=10, trim=True)
    if show:
        plt.show()
    return fig, ax

def plot_model_vs_data(Polynomial, ax=None, sample_data=None, metric='adjusted_r2', show=True):
    """
    Plots the polynomial approximation against the true data.

    :param Polynomial Polynomial: 
        An instance of the Poly class.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param list sample_data:
        A list formed by ``[X, y]`` where ``X`` represents the spatial data input and ``y`` the output.
    :param bool show: 
        Option to view the plot.
    """
    if ax is None:
        fig,ax = plt.subplots(figsize=(8, 6),tight_layout=True)
        ax.set_xlabel('Polynomial model')
        ax.set_ylabel('True data')
    else:
        fig = ax.figure
    if sample_data is None:
        X = Polynomial.get_points()
        y_truth = Polynomial._model_evaluations
        y_model = Polynomial.get_polyfit(X)
    else:
        X, y_truth = sample_data[0], sample_data[1]
        y_model = Polynomial.get_polyfit(X)
    myscore = score(y_truth, y_model, metric, X)
    ax.plot(y_model, y_truth, 'o', color='dodgerblue', ms=10, markeredgecolor='k',lw=1, alpha=0.6)
    displaytext = '$Score$ = '+str(np.round(float(myscore), 2))
    ax.text(0.3, 0.9, displaytext, transform=ax.transAxes, \
        horizontalalignment='center', verticalalignment='center', fontsize=14, color='grey')
    sns.despine(offset=10, trim=True)
    if show:
        plt.show()
    return fig, ax

def plot_decision_surface(PolyTree,ij,ax=None,X=None,y=None,max_depth=None,label=True,
                                 color='data',colorbar=True,show=True,kwargs={}):
    """
    Plots the decision boundaries of the PolyTree over a 2D surface.

    :param PolyTree PolyTree: 
        An instance of the PolyTree class.
    :param list ij: 
        A list containing the two dimensions to plot over. For example, ``ij=[6,7]`` with plot over the 6th and 7th dimensions in ``X``.
    :param matplotlib.ax ax: 
        An instance of the ``matplotlib`` axes class to plot onto. If ``None``, a new figure and axes are created (default: ``None``).
    :param :numpy.ndarray X:
        A numpy ndarray containing the input data to plot.
    :param :numpy.ndarray y:
        A numpy ndarray containing the output data to plot.
    :param int max_depth:
        The maximum tree depth to plot decision boundaries for.
    :param bool label:
        If ``True`` then the subdomains are labelled by their node number.
    :param string color:
        What to color the scatter points by. ``'data'`` to color by the ``X``,``y`` data. ``'predict'`` to color by the PolyTree predictions, and ``'error'`` to color by the predictive error. (default: ``'data'``).
    :param bool colorbar:
        Option to add a colorbar.
    :param bool show:
        Option to view the plot.
    :param dict kwargs:
        Dictionary of keyword arguments to pass to matplotlib.scatter().  
    """
    # Set default kwargs
    kwargs = set_defaults(kwargs, {'alpha':0.8,'ec':'lightgray','cmap':'coolwarm'})

    if ax is None:
        fig, ax = plt.subplots(tight_layout=True)
    else:
        fig = ax.figure

    # TODO - What to do in case where "data" doesn't exist? Have domain\subdomain attribute for each node?
    if X is None:
            X = PolyTree.tree["data"][0]
    if y is None:
            y = PolyTree.tree["data"][1]
    Xij = X[:,ij]
    PolyTree.tree["Xmin"] = np.min(Xij,axis=0)
    PolyTree.tree["Xmax"] = np.max(Xij,axis=0)

    if color.lower() == 'error':
            error = y - PolyTree.predict(X)
            scat = ax.scatter(X[:,ij[0]],X[:,ij[1]],c=error,**kwargs)
            label = 'Error'
    elif color.lower() == 'predict':
            scat = ax.scatter(X[:,ij[0]],X[:,ij[1]],c=PolyTree.predict(X),**kwargs)
            label = 'Prediction'
    elif color.lower() == 'data':
            scat = ax.scatter(X[:,ij[0]],X[:,ij[1]],c=y,**kwargs)
            label = 'Truth'
    else:
        raise ValueError("color argument should be set to 'error', 'predict', or 'data'")
    if colorbar:
        fig.colorbar(scat, orientation="vertical", pad=0.1, shrink=0.5,label=label)

    def _get_boundaries(nodes,final):
            # Find leaf nodes
            left_children = [node["children"]["left"] for node in nodes]
            leaf_nodes = np.array([True if node is None else False for node in left_children])

            # Get splitting info from non-leaf nodes (i.e. split nodes)
            split_nodes = nodes[~leaf_nodes]
            split_dims = [node["j_feature"] for node in split_nodes]
            split_vals = [node["threshold"] for node in split_nodes]
            indices    = [node["index"]     for node in split_nodes]

            # Labelling done before splits, as we only label up to max_depth and then return
            if label:
                    # If final, label all nodes, else only leaf nodes
                    if final:
                            for node in nodes:
                                    ax.annotate('Node %d'%node["index"],(node["Xmax"][0],node["Xmax"][1]),
                                                ha='right',va='top',textcoords='offset points',
                                                xytext=(-5, -5))
                            return
                    else:
                            for node in nodes[leaf_nodes]:
                                    ax.annotate('Node %d'%node["index"],(node["Xmax"][0],node["Xmax"][1]),
                                                ha='right',va='top',textcoords='offset points',
                                                xytext=(-5, -5))

            # Plot split lines
            for n, node in enumerate(split_nodes):
                    if split_dims[n]==ij[0]:
                            ax.vlines(split_vals[n],node["Xmin"][1],
                                       node["Xmax"][1],'k')
                    else:
                            ax.hlines(split_vals[n],node["Xmin"][0],
                                       node["Xmax"][0],'k')

            # Update bounding boxes of child nodes before returning them
            for node in split_nodes:
                    if node["j_feature"]==ij[0]:
                            node["children"]["left"]["Xmax"]  = [node["threshold"],node["Xmax"][1]]
                            node["children"]["right"]["Xmin"] = [node["threshold"],node["Xmin"][1]]
                            node["children"]["left"]["Xmin"]  = node["Xmin"]
                            node["children"]["right"]["Xmax"] = node["Xmax"]
                    else:
                            node["children"]["left"]["Xmax"]  = [node["Xmax"][0],node["threshold"]]
                            node["children"]["right"]["Xmin"] = [node["Xmin"][0],node["threshold"]]
                            node["children"]["left"]["Xmin"]  = node["Xmin"]
                            node["children"]["right"]["Xmax"] = node["Xmax"]

            # Extract child node info for next level down
            left_nodes  = [node["children"]["left"]  for node in split_nodes]
            right_nodes = [node["children"]["right"] for node in split_nodes]

            child_nodes = np.array(left_nodes + right_nodes)

            return child_nodes

    nodes = np.array([PolyTree.tree])
    depth = 0
    final = False
    while len(nodes)>0:
            if depth==max_depth: final = True
            nodes = _get_boundaries(nodes,final)
            if final: break
            depth += 1

    if show:
        plt.show()
    return fig, ax, scat

def set_defaults(kwargs, defaults):
    for key in defaults:
        kwargs.setdefault(key, defaults[key])
    return kwargs



""" Utilities for dealing with correlated inputs."""
from equadratures.parameter import Parameter
from equadratures.poly import Poly, evaluate_model, evaluate_model_gradients
from equadratures.basis import Basis
import numpy as np
from scipy import stats
from scipy.stats import multivariate_normal, norm
from scipy.linalg import lu
from scipy import optimize
from copy import deepcopy

class Correlations(object):
    """
    The class defines methods for polynomial approximations with correlated inputs, including the Nataf transform and Gram-Schmidt process.

    :param numpy.ndarray correlation_matrix: The correlation matrix associated with the joint distribution.
    :param Poly poly: Polynomial defined with parameters with marginal distributions in uncorrelated space.
    :param list parameters: List of parameters with marginal distributions.
    :param str method: `nataf-transform` or `gram-schmidt`.
    :param bool verbose: Display Cholesky decomposition of the fictive matrix.

    **Example usage**::

        def func(x):
            s1 = s[0]
            s2 = s[1]
            return s1**2 - s2**3 - 2. * s1 - 0.5 * s2

        parameters = [Parameter(distribution='beta', shape_parameter_A=5.0, shape_parameter_B=2.0, lower=0.0, upper=1.0, order=15) for _ in range(2)]
        basis = Basis('total-order')
        uncorr_poly = Poly(parameters, basis, method='least-squares',
                            sampling_args={'mesh': 'monte-carlo',
                                            'subsampling-algorithm': 'lu'})

        corr_mat = np.array([[1.0, -0.855],
                            [-0.855, 1.0]])

        corr_gs = Correlations(corr_mat, poly=uncorr_poly, method='gram-schmidt')
        corr_gs.set_model(func)
        corrected_poly_gs = corr_gs.get_transformed_poly()
        print(corrected_poly_gs.get_mean_and_variance())

    **References**
        1. Melchers, R. E., (1945) Structural Reliability Analysis and Predictions. John Wiley and Sons, second edition.
        2. Jakeman, J. D. et al., (2019) Polynomial chaos expansions for dependent random variables.
    """
    def __init__(self, correlation_matrix, poly=None, parameters=None, method=None, verbose=False):
        if (poly is None) and (method is not None):
            raise ValueError('Need to specify poly for probability transform.')
        if poly is not None:
            self.poly = poly
            D = self.poly.get_parameters()
        elif parameters is not None:
            D = parameters
        else:
            raise ValueError('Need to specify either poly or parameters.')
        self.D = D
        self.R = correlation_matrix
        self.std = Parameter(order=5, distribution='normal',shape_parameter_A = 0.0, shape_parameter_B = 1.0)
        inf_lim = -8.0
        sup_lim = - inf_lim
        p1 = Parameter(distribution = 'uniform', lower = inf_lim, upper = sup_lim, order = 31)
        myBasis = Basis('tensor-grid')
        self.Pols = Poly([p1, p1], myBasis, method='numerical-integration')
        Pols = self.Pols
        p = Pols.get_points()
        # w = Pols.get_weights()
        w = Pols.get_weights() * (sup_lim - inf_lim)**2
        p1 = p[:,0]
        p2 = p[:,1]
        R0 = np.eye((len(self.D)))
        for i in range(len(self.D)):
            for j in range(i+1, len(self.D), 1):
                if self.R[i,j] == 0:
                    R0[i,j] = 0.0
                else:
                    z1 = np.array(self.D[i].get_icdf(self.std.get_cdf(points=p1)))
                    z2 = np.array(self.D[j].get_icdf(self.std.get_cdf(points=p2)))

                    tp11 = (z1 - self.D[i].mean) / np.sqrt( self.D[i].variance )
                    tp22 = (z2 - self.D[j].mean)/ np.sqrt( self.D[j].variance )

                    coefficientsIntegral = np.flipud(tp11*tp22 * w)
                    def check_difference(rho_ij):
                        bivariateNormalPDF = (1.0 / (2.0 * np.pi * np.sqrt(1.0-rho_ij**2)) * np.exp(-1.0/(2.0*(1.0 - rho_ij**2)) * (p1**2 - 2.0 * rho_ij * p1 * p2  + p2**2 )))
                        diff = np.dot(coefficientsIntegral, bivariateNormalPDF)
                        return diff - self.R[i,j]

                    # if (self.D[i].name!='custom') or (self.D[j].name!='custom'):
                    rho = optimize.newton(check_difference, self.R[i,j], maxiter=50)
                    # else:
                    #     # ???
                    #     res = optimize.least_squares(check_difference, self.R[i,j], bounds=(-0.999,0.999), ftol=1.e-03)
                    #     rho = res.x
                    #     print('A Custom Marginal is present')

                    R0[i,j] = rho
                    R0[j,i] = R0[i,j]
        self.R0 = R0.copy()

        self.A = np.linalg.cholesky(R0)
        if verbose:
            print('The Cholesky decomposition of fictive matrix R0 is:')
            print(self.A)
            print('The fictive matrix is:')
            print(R0)

        if method is None:
            pass
        elif method.lower() == 'nataf-transform':
            list_of_parameters = []
            for i in range(0, len(self.D)):
                standard_parameter = Parameter(order=self.D[i].order, distribution='gaussian', shape_parameter_A = 0., shape_parameter_B = 1.)
                list_of_parameters.append(standard_parameter)

            # have option so that we don't need to obtain
            self.corrected_poly = deepcopy(self.poly)

            if hasattr(self.corrected_poly, '_quadrature_points'):
                self.corrected_poly._set_parameters(list_of_parameters)
                self.standard_samples = self.corrected_poly._quadrature_points
                self._points = self.get_correlated_samples(X=self.standard_samples)
                # self.corrected_poly._quadrature_points = self._points.copy()
        elif method.lower() == 'gram-schmidt':
            basis_card = poly.basis.cardinality
            oversampling = 10

            N_Psi = oversampling * basis_card
            S_samples = self.get_correlated_samples(N=N_Psi)
            w_weights = 1.0 / N_Psi * np.ones(N_Psi)
            Psi = poly.get_poly(S_samples).T
            WPsi = np.diag(np.sqrt(w_weights)) @ Psi
            self.WPsi = WPsi

            R_Psi = np.linalg.qr(WPsi)[1]

            self.R_Psi = R_Psi
            self.R_Psi[0, :] *= np.sign(self.R_Psi[0, 0])
            self.corrected_poly = deepcopy(poly)
            self.corrected_poly.inv_R_Psi = np.linalg.inv(self.R_Psi)
            self.corrected_poly.corr = self
            self.corrected_poly._set_points_and_weights()

            P = self.corrected_poly.get_poly(self.corrected_poly._quadrature_points)
            W = np.mat(np.diag(np.sqrt(self.corrected_poly._quadrature_weights)))
            A = W * P.T
            self.corrected_poly.A = A
            self.corrected_poly.P = P

            if hasattr(self.corrected_poly, '_quadrature_points'):
                # TODO: Correlated quadrature points?
                self._points = self.corrected_poly._quadrature_points
        else:
            raise ValueError('Invalid method for correlations.')
    def get_points(self):
        """
        Returns the quadrature points accounting for correlations. For Nataf transform, returns points in the correlated standard normal space.
        For Gram-Schmidt, returns points according to the GS polynomial basis.

        :param Correlations self: An instance of the Correlations object.

        :return:
            **points**: A numpy.ndarray of quadrature points with shape (number_of_samples, dimension).

        """
        return self._points
    def set_model(self, model=None, model_grads=None):
        """
        Computes the coefficients of transformed polynomial (equivalent to calling `self.get_transformed_poly().set_model(...)`)

        :param Correlations self:
            An instance of the Correlations class.
        :param callable model:
            The function that needs to be approximated. In the absence of a callable function, the input can be the function evaluated at the quadrature points.
        :param callable model_grads:
            The gradient of the function that needs to be approximated. In the absence of a callable gradient function, the input can be a matrix of gradient evaluations at the quadrature points.
        """
        # Need to account for the nataf transform here?
        model_values = None
        model_grads_values = None
        if callable(model):
            model_values = evaluate_model(self._points, model)
        else:
            model_values = model
        if model_grads is not None:
            if callable(model_grads):
                model_grads_values = evaluate_model_gradients(self._points, model_grads)
            else:
                model_grads_values = model_grads
        self.corrected_poly.set_model(model_values, model_grads_values)
    def get_transformed_poly(self):
        """
        Returns the transformed polynomial.

        :param Correlations self:
            An instance of the Correlations class.

        :return:
            **poly**: An instance of the Poly class.
        """
        return self.corrected_poly
    def get_correlated_samples(self, N=None, X=None):
        """
        Method for generating correlated samples.

        :param int N: 
            Number of correlated samples required.
        :param ndarray X: 
            (Optional) Points in the uncorrelated space to map to the correlated space.

        :return:
            **C**: A numpy.ndarray of shape (N, M), which contains the correlated samples.
        """
        d = len(self.D)

        if X is None:
            if N is None:
                raise ValueError('Need to specify number of points to generate.')
            X = np.random.multivariate_normal(np.zeros(d), np.eye(d), size=N)

        X_test = X @ self.A.T

        U_test = np.zeros(X_test.shape)
        for i in range(d):
            U_test[:, i] = self.std.get_cdf(X_test[:, i])

        Z_test = np.zeros(U_test.shape)
        for i in range(d):
            Z_test[:, i] = self.D[i].get_icdf(U_test[:, i])
        return Z_test

    def get_pdf(self, X):
        """
        Evaluate PDF at the sample points.

        :param numpy.ndarray X: 
            Sample points (Number of points by dimensions)

        :return:
            **C**: A numpy.ndarray of shape (N,) with evaluations of the PDF.
        """
        parameters = self.D
        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        d = X.shape[1]

        U = np.zeros(X.shape)
        for i in range(d):
            U[:, i] = norm.ppf(parameters[i].get_cdf(X[:, i]))
        cop_num = multivariate_normal(mean=np.zeros(d), cov=self.R0).pdf(U)
        cop_den = np.prod(np.array([norm.pdf(U[:, i]) for i in range(d)]), axis=0)
        marginal_prod = np.prod(np.array([parameters[i].get_pdf(X[:, i]) for i in range(d)]), axis=0)
        return cop_num / cop_den * marginal_prod
